/*	Copyright (C) 2003-2015 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#include "openil.pmh"

namespace fe
{

ImageIL::ImageIL(void)
{
	ilInit();
	iluInit();
//	ilutInit();
//	ilutRenderer(ILUT_OPENGL);

	iluImageParameter(ILU_FILTER,ILU_SCALE_BELL);	//* arbitrary
	ilEnable(IL_FILE_OVERWRITE);
//	ilEnable(IL_ORIGIN_SET);
//	ilOriginFunc(IL_ORIGIN_LOWER_LEFT);
}

ImageIL::~ImageIL(void)
{
}

U32 ImageIL::createSelect(void)
{
	ILuint id;
	ilGenImages(1,&id);
	ilBindImage(id);

	return (U32)id;
}

U32 ImageIL::loadSelect(String filename)
{
	ILuint id;
	ilGenImages(1,&id);
	ilBindImage(id);
	if(!ilLoadImage(const_cast<char*>(filename.c_str())))
	{
		ilDeleteImages(1,(ILuint*)&id);
		return 0L;
	}

	return (U32)id;
}

U32 ImageIL::interpretSelect(String a_source)
{
	//* would this ever make sense?
	return interpretSelect(const_cast<char*>(a_source.c_str()),
			a_source.length());
}

U32 ImageIL::interpretSelect(void* data,U32 size)
{
	ILuint id;
	ilGenImages(1,&id);
	ilBindImage(id);
	if(!ilLoadL(IL_TYPE_UNKNOWN,data,size))
	{
		ilDeleteImages(1,(ILuint*)&id);
		return 0L;
	}

	return (U32)id;
}

BWORD ImageIL::save(String filename)
{
	ILboolean success=ilSaveImage(const_cast<char*>(filename.c_str()));
	return success!=0;
}

void ImageIL::select(U32 id)
{
	ilBindImage((ILuint)id);
}

void ImageIL::unload(U32 id)
{
	ilDeleteImages(1,(ILuint*)&id);
}

void ImageIL::setFormat(ImageI::Format format)
{
	ILenum formatid=IL_COLOUR_INDEX;
	switch(format)
	{
		case e_rgb:
			formatid=IL_RGB;
			break;
		case e_rgba:
			formatid=IL_RGBA;
			break;
		default:
			break;
	}

//	feLog("ImageIL::setFormat %d %d %d %d %d %d\n",width(),height(),depth(),
//			bppID(formatid),formatid,typeID(formatid));

	ILboolean success=ilTexImage(width(),height(),depth(),
			bppID(formatid),formatid,typeID(formatid),NULL);

	if(!success)
	{
		feX("ImageIL::setFormat","ilTexImage failed");
	}
}

void ImageIL::resize(U32 width,U32 height,U32 depth)
{
#ifdef __x86_64__

#if FE_CODEGEN<=FE_DEBUG
	feLog("ImageIL::resize"
			" WARNING, x86_64 OpenIL workaround discards old image\n");
#endif

//	feLog("ImageIL::resize %d %d %d %d %d %d\n",width,height,depth,
//			bppID(formatID()),formatID(),typeID(formatID()));

	ILboolean success=ilTexImage(width,height,depth,
			bppID(formatID()),formatID(),typeID(formatID()),NULL);

	if(!success)
	{
		feX("ImageIL::resize","ilTexImage failed");
	}
#else
	iluScale((ILuint)width,(ILuint)height,(ILuint)depth);
#endif
}

void ImageIL::replaceRegion(U32 x,U32 y,U32 z,
		U32 width,U32 height,U32 depth,void* data)
{
	ilSetPixels(x,y,z,width,height,depth,formatID(),typeID(formatID()),data);
}

ImageI::Format ImageIL::format(void) const
{
	switch(formatID())
	{
		case IL_COLOUR_INDEX:
			return e_colorindex;
		case IL_RGB:
			return e_rgb;
		case IL_RGBA:
			return e_rgba;
	}
	return e_none;
}

U32 ImageIL::width(void) const
{
	return ilGetInteger(IL_IMAGE_WIDTH);
}

U32 ImageIL::height(void) const
{
	return ilGetInteger(IL_IMAGE_HEIGHT);
}

U32 ImageIL::depth(void) const
{
	return ilGetInteger(IL_IMAGE_DEPTH);
}

void* ImageIL::raw(void) const
{
	return ilGetData();
}

} // namespace
