/*	Copyright (C) 2003-2015 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#ifndef __operator_HingeOp_h__
#define __operator_HingeOp_h__

namespace fe
{

/**************************************************************************//**
	@brief Operator to rotate a surface out of another

	@ingroup operator
*//***************************************************************************/
class FE_DL_EXPORT HingeOp:
	public OperatorThreaded,
	public Initialize<HingeOp>
{
	public:

					HingeOp(void):
						m_frame(-1)										{}
virtual				~HingeOp(void)										{}

		void		initialize(void);

					//* As HandlerI
virtual	void		handle(Record& a_rSignal);
virtual	void		run(I32 a_id,sp<SpannedRange> a_spRange);

	private:
static	void		calcConversion(SpatialTransform& a_rTransform,
							const Real a_angle,const SpatialVector& a_rAxis,
							const SpatialVector& a_rPivot);

		I32							m_frame;

		sp<SurfaceAccessibleI>		m_spOutputAccessible;
		sp<SurfaceI>				m_spObstacle;

		Array<Real>					m_totalAngle;
		Array<Real>					m_pointAngle;
		Array<Real>					m_maxAngle;

		SpatialVector				m_pivotA;
		SpatialVector				m_pivotB;
		SpatialVector				m_pivotC;

		I32							m_partitionIndex;

		sp<DrawI>					m_spDrawGuide;

		sp<Profiler>				m_spProfiler;
		sp<Profiler::Profile>		m_spProfilePrep;
		sp<Profiler::Profile>		m_spProfileLoop;
		sp<Profiler::Profile>		m_spProfilePush;
		sp<Profiler::Profile>		m_spProfilePull;
		sp<Profiler::Profile>		m_spProfileRotate;
};

} /* namespace */

#endif /* __operator_HingeOp_h__ */

